# Standard Library imports
from __future__ import annotations
from typing import cast, TYPE_CHECKING
from textwrap import shorten

if TYPE_CHECKING:
    from textualdon.toot import TootWidget

# Textual imports
from textual import on
from textual.message import Message
from textual.binding import Binding
from textual.screen import ModalScreen
from textual.containers import Horizontal, Vertical, Container
from textual.widgets import Label, Button

# TextualDon imports
from textualdon.simplebutton import SimpleButton
from textualdon.messages import ExamineToot, UserPopupMessage
from textualdon.screens import NotImplementedScreen, TextualdonModalScreen
from textualdon.sql import SQLite
from textualdon.bs4_parser import BS4Parser



class TootOptionsScreen(ModalScreen):

    class Dismiss(Message):
        pass

    BINDINGS = [
        Binding("escape", "dismiss", key_display='Esc', description="Close the pop-up screen.", show=True),
        Binding("up", "focus_previous", description="Focus the previous button."),
        Binding("down", "focus_next", description="Focus the next button."),
    ]

    controls = "Arrow keys (or tab): navigate | Enter: select |  Esc or click anywhere: close."

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.db = cast(SQLite, self.app.sqlite)

        sql_query = "SELECT value FROM settings WHERE name = ?"
        row1 = self.db.fetchone(sql_query, ("view_json_active",))
        self.view_json_active: bool = ('True' == row1[0])

    def on_mount(self):
        
        options_container = self.query_one("#options_container")
        options_list = []

        # NOTE: this might seem like kind of a mess, but they have to go in a specific order.
        # So its either this or insert them in the right order, which is more error-prone.

        if self.toot_widget.toot_content_container.image_on:
            options_list.append(SimpleButton(    'View image in full size', id='view_image_button', classes="screen_button"))
        options_list.append(SimpleButton(  'Copy link to toot to clipboard', id='copy_link_button', classes="screen_button"))
        if self.toot_widget.card and self.toot_widget.card['url']:
            options_list.append(SimpleButton("Open mentioned URL / Website.", id="open_url_button", classes="screen_button"))
        if self.toot_widget.in_reply_to_id:
            options_list.append(SimpleButton(        "View conversation", id="view_parent_button",  classes="screen_button"))
        options_list.append(SimpleButton(       'View details of author', id='view_profile_button', classes="screen_button"))
        if self.toot_widget.boosted_by:
            options_list.append(SimpleButton(  "View details of booster", id="view_booster_button", classes="screen_button"))
        if self.view_json_active:
            if self.toot_widget.toot_switcher.current == "toot_json_container":
                options_list.append(SimpleButton(  "Back to normal mode", id="toggle_json_button",  classes="screen_button"))
            else:
                options_list.append(SimpleButton(            "View JSON", id="toggle_json_button",  classes="screen_button"))
        options_list.append(SimpleButton('Refresh Toot', id='refresh_button', classes="screen_button"))

        options_container.mount_all(options_list)

        self.mount(
            Container(Label(self.controls, classes='screen_label'), classes='screen_container wide help')
        )     

    # NOTE: Every single thing a person can click or press enter on here should result in the
    # options screen being dismissed. So here that behavior is centralized.

    def on_click(self):
        self.dismiss()

    @on(SimpleButton.Pressed, selector="#view_image_button")
    async def view_image(self) -> None:
        await self.toot_widget.view_image()

    @on(SimpleButton.Pressed, selector="#copy_link_button")
    def copy_link(self) -> None:
        self.app.copy_to_clipboard(self.toot_widget.toot_url)

    @on(SimpleButton.Pressed, selector="#open_url_button")
    def open_url(self) -> None:
        self.app.handle_link(self.toot_widget.card['url'])

    @on(SimpleButton.Pressed, selector="#view_parent_button")
    def view_parent(self) -> None:
        self.post_message(ExamineToot(self.toot_widget.in_reply_to_id))

    @on(SimpleButton.Pressed, selector="#view_profile_button")
    def view_profile(self) -> None:
        self.toot_widget.open_user_popup()

    @on(SimpleButton.Pressed, selector="#view_booster_button")
    def view_booster(self) -> None:
        self.toot_widget.open_booster_popup()

    @on(SimpleButton.Pressed, selector="#refresh_button")
    def refresh_toot(self) -> None:
        self.toot_widget.refresh_toot()
        
    def action_focus_previous(self):
        self.focus_previous()
    
    def action_focus_next(self):
        self.focus_next()

    @on(SimpleButton.Pressed, selector="#toggle_json_button")
    def toggle_json(self) -> None:

        if self.toot_widget.toot_switcher.current == "toot_json_container":
            self.toot_widget.switch_content("toot_content_container")
        else:
            self.toot_widget.switch_content("toot_json_container")


class TootOptionsMainUser(TootOptionsScreen):

    def __init__(self, toot_widget: TootWidget, **kwargs):
        super().__init__(**kwargs)
        self.toot_widget = toot_widget
        self.option_handler = toot_widget.option_handler

        if self.toot_widget.is_main_user:
            self.pin_status = self.toot_widget.json["pinned"]       #! must refresh between toots?
            self.pin_label = "Unpin from Profile" if self.pin_status else "Pin on Profile"
            self.mute_status = self.toot_widget.json["muted"]
            self.mute_label = "Unmute conversation" if self.mute_status else "Mute conversation"

    def compose(self):
        with Vertical(id="options_container", classes='screen_container'):
            yield SimpleButton(self.pin_label, id="pin_button", classes="screen_button")
            yield SimpleButton("Edit",   id="edit_button",      classes="screen_button")
            yield SimpleButton("Delete", id="delete_button",    classes="screen_button")
            yield SimpleButton("Delete & Redraft", id="delete_redraft_button", classes="screen_button")

            yield SimpleButton(self.mute_label, id='mute_button', classes="screen_button")  #! exists between other user? 

    # This is identical in both classes but it has to be on the actual instance and not
    # the parent class. Don't fully understand why. But it didn't work when it was in the parent class.
    @on(SimpleButton.Pressed)
    def close_screen(self):
        self.dismiss()

    # NOTE: The options are handled by the TootOptionsHandler class, attached to the TootWidget.

    @on(SimpleButton.Pressed, selector="#pin_button")
    async def pin_toot(self) -> None:
        action          = "status_unpin"   if self.pin_status else "status_pin"
        success_message = "Toot unpinned." if self.pin_status else "Toot pinned."
        failure_message = "FAILED to unpin toot." if self.pin_status else "FAILED to pin toot."

        await self.toot_widget.option_handler.handle_toot_action(
            action=action,
            success_message=success_message,
            failure_message=failure_message,
            toggle_status=self.pin_status,
            key=self.option_handler.StatusKey.PINNED,
        )
        
    @on(SimpleButton.Pressed, selector="#mute_button")
    async def mute_toot(self) -> None:
        action          = "status_unmute"     if self.mute_status else "status_mute"
        success_message = "Toot unmuted."     if self.mute_status else "Toot muted."
        failure_message = "FAILED to unmute toot." if self.mute_status else "FAILED to mute toot."

        await self.toot_widget.option_handler.handle_toot_action(
            action=action,
            success_message=success_message,
            failure_message=failure_message,
            toggle_status=self.mute_status,
            key=self.option_handler.StatusKey.MUTED,
        )
        
    @on(SimpleButton.Pressed, selector="#delete_button")
    def delete_toot_button(self) -> None:

        # NOTE: Because some methods like here require a callback from a confirmation 
        # screen, they need to be handled by the toot widget itself. 

        self.toot_widget.post_message(self.toot_widget.DeleteToot())
        
    @on(SimpleButton.Pressed, selector="#delete_redraft_button")
    def delete_redraft_toot(self) -> None:

        self.toot_widget.post_message(self.toot_widget.DeleteToot(redraft=True)) 
        
    @on(SimpleButton.Pressed, selector="#edit_button")
    def edit_toot(self) -> None:
        self.toot_widget.edit_toot()     # toot_widget contains the text edit box.
        

class TootOptionsOtherUser(TootOptionsScreen):

    def __init__(self, toot_widget: TootWidget, **kwargs):
        super().__init__(**kwargs)
        self.toot_widget = toot_widget
        # self.option_handler = toot_widget.option_handler

        self.toot_username = toot_widget.username
        self.mention_label = f"Mention @{self.toot_username}"

        self.mute_status = self.toot_widget.relation_dict['muting']
        self.mute_label = f"Unmute @{self.toot_username}" if self.mute_status else f"Mute @{self.toot_username}"

        self.block_status = self.toot_widget.relation_dict['blocking']
        self.block_label = f"Unblock @{self.toot_username}" if self.block_status else f"Block @{self.toot_username}"

        self.report_label = f"Report @{self.toot_username}"

    def compose(self):
        with Vertical(id="options_container", classes='screen_container'):
            yield SimpleButton(self.mention_label, id="mention_button", classes="screen_button")
            yield SimpleButton(self.mute_label,    id="mute_button",    classes="screen_button")
            yield SimpleButton(self.block_label,   id="block_button",   classes="screen_button")
            yield SimpleButton(self.report_label,  id="report_button",  classes="screen_button")
            yield SimpleButton("Filter post",      id="filter_button",  classes="screen_button")

    @on(SimpleButton.Pressed)
    def close_screen(self):
        self.dismiss()   

    @on(SimpleButton.Pressed, selector="#mention_button")
    def mention_handler(self):
        self.app.main_tootbox.set_text(f"@{self.toot_username} ")
        
    @on(SimpleButton.Pressed, selector="#mute_button")
    def mute_button_handler(self):

        self.toot_widget.post_message(self.toot_widget.MuteUser())
            
    @on(SimpleButton.Pressed, selector="#block_button")
    def block_button_handler(self):

        self.toot_widget.post_message(self.toot_widget.BlockUser())
        
    @on(SimpleButton.Pressed, selector="#report_button")
    def report_user(self) -> None:
        self.app.push_screen(NotImplementedScreen('Reporting system', classes="modal_screen"))

    @on(SimpleButton.Pressed, selector="#filter_button")
    def filter_toot(self) -> None:
        self.app.push_screen(NotImplementedScreen('Filter toot', classes="modal_screen"))
        

class MuteScreen(TextualdonModalScreen):
    """Called by:
    - TootWidget.mute_user
    
    Callback: TootWidget._mute_user"""

    def __init__(self, username: str, mute_status: bool, **kwargs):
        self.username = username
        self.mute_status = mute_status
        super().__init__(**kwargs)        

    def compose(self):
        with Container(classes='screen_container'):
            yield Label(f'Are you sure you want to mute @{self.username}? \n', classes='screen_label')
            with Horizontal(classes='screen_buttonbar'):
                yield Button('Yes', id='confirm_yes_button')                
                yield Button('No', id='close_button')           

    @on(Button.Pressed, selector='#confirm_yes_button')
    def confirm_yes(self):

        self.dismiss(self.mute_status)

    @on(Button.Pressed, selector='#close_button')
    def report_close(self):
        self.action_pop_screen()


class BlockScreen(TextualdonModalScreen):
    """Called by:
    - TootWidget.block_user
    
    Callback: TootWidget._block_user"""

    def __init__(self, username: str, block_status: bool, **kwargs):
        self.username = username
        self.block_status = block_status
        super().__init__(**kwargs)

    def compose(self):
        with Container(classes='screen_container'):
            yield Label(f'Are you sure you want to block @{self.username}? \n', classes='screen_label')
            with Horizontal(classes='screen_buttonbar'):
                yield Button('Yes', id='confirm_yes_button')                
                yield Button('No', id='close_button')            

    @on(Button.Pressed, selector='#confirm_yes_button')
    def confirm_yes(self):

        self.dismiss(self.block_status)

    @on(Button.Pressed, selector='#close_button')
    def report_close(self):

        self.action_pop_screen()


class UserPopup(TextualdonModalScreen):
    """ Called by:
    - TootWidget.open_user_popup
    - TootWidget.open_booster_popup
    
    Callback: None (sends UserPopupMessage instead)"""

    controls = "Arrow keys (or tab): navigate | Enter: select | Esc or click anywhere: close"
    
    bs4_parser = BS4Parser()

    def __init__(self, account_dict: dict, relation_dict: dict, **kwargs):
        super().__init__(**kwargs)
        self.account = account_dict
        self.relation = relation_dict
        self.parsed_bio = self.bs4_parser.parser(self.account["note"])
        self.disable_follow = False

        if self.account['id'] == self.app.logged_in_user_id:
            self.relation_note1 = 'This is you.'
            self.relation_note2 = ''
            self.disable_follow = True
            return

        if self.relation['following']:
            self.relation_note1 = 'You [green]are following[/green] this user.'
        else:
            self.relation_note1 = 'You are [red]not following[/red] this user.'
        if self.relation['followed_by']:
            self.relation_note2 = 'They [green]are following[/green] you \n'
        else:
            self.relation_note2 = 'They are [red]not following[/red] you \n'

    def compose(self):

        user_label = (
            f"{self.account['display_name']}\n"
            f"@{self.account['acct']}\n\n"
            f"{self.account['followers_count']} Followers \n\n"
            f"{shorten(self.parsed_bio, width=80)}\n\n"
            f"{self.relation_note1}\n"
            f"{self.relation_note2}"
        )
        with Container(classes="screen_container"):
            yield Label(user_label, classes='screen_label')
            with Horizontal(classes='screen_buttonbar'):
                yield SimpleButton(
                    'Follow',
                    id='user_follow_button',
                    classes='screen_button',
                    disabled=self.disable_follow
                )
                yield SimpleButton(
                    'Go to Profile',
                    id='go_to_profile_button',
                    classes='screen_button',
                )             

    def on_mount(self):

        self.follow_button = self.query_one('#user_follow_button')
        self.profile_button = self.query_one('#go_to_profile_button')

        if self.relation['following']:
            self.follow_button.update('Unfollow')
        if self.disable_follow:
            self.follow_button.visible = False
    
    def on_click(self):
        self.dismiss()

    @on(SimpleButton.Pressed, selector='#user_follow_button')
    def follow_user(self):
        self.post_message(UserPopupMessage('follow', self.account, self.relation))
        self.dismiss()

        # NOTE: Message is handled in the main app.

    @on(SimpleButton.Pressed, selector='#go_to_profile_button')
    def go_to_profile(self):
        self.post_message(UserPopupMessage('profile', self.account, self.relation))
        self.dismiss()    



# Standard Library Imports
from __future__ import annotations
from typing import cast
import uuid

# Textual Imports
from textual import on
from textual.dom import NoScreen
from textual.message import Message
from textual.binding import Binding
from textual.containers import Horizontal, Container  # VerticalScroll, Grid
from textual.widget import Widget
from textual.widgets import (
    TextArea,
)

# TextualDon Imports
from textualdon.simplebutton import SimpleButton
from textualdon.messages import SuperNotify, ScrollToWidget
from textualdon.screens import NotImplementedScreen
from textualdon.messages import SwitchMainContent


class TextAreaMain(TextArea):

    class Submit(Message):
        pass

    class Search(Message):
        pass

    class Hide(Message):
        pass

    BINDINGS = [
        Binding("ctrl+e", "submit", "Submit", show=True),
        Binding("ctrl+f", "search", "Search", show=True),
        Binding("ctrl+b", "hide", "Hide", show=True),
        Binding(key="f6", action="bookmarks", description="Bookmarks", key_display="F6", show=False),
        Binding(key="f7", action="favorites", description="Favorites", key_display="F7", show=False),
    ]
    
    def action_submit(self):
        self.post_message(self.Submit())

    def action_search(self):
        self.post_message(self.Search())

    def action_hide(self):
        self.screen.focus_next()
        self.post_message(self.Hide())

    # Overwriting F6 and F7 is a really hacky way to solve this problem. I'd prefer to be able
    # to remove the bindings from the list. But I can't find a way to do that.

    def action_bookmarks(self):
        self.post_message(SwitchMainContent("bookmarks"))

    def action_favorites(self):
        self.post_message(SwitchMainContent("favorites"))

    def on_focus(self):
        self.action_cursor_line_end()


class TextAreaReply(TextArea):

    class Submit(Message):
        pass

    BINDINGS = [
        Binding("ctrl+e", "submit", "Submit reply", show=True),
        Binding("escape", "cancel", "Cancel reply", key_display='Esc', show=True),
        # Binding("enter", "pass", show=False),
    ]

    def on_mount(self):
        self.focus()

    def focus(self, scroll_visible: bool = True):
        """Copied from Widget.focus() but with scroll_visible set to False."""

        try:
            self.screen.set_focus(self, scroll_visible=False)
        except NoScreen:
            pass

        self.post_message(ScrollToWidget(self))
        self.parent.parent.toot_widget.on_focus()

        # This is the original code from Widget.focus(). I'm not sure why it's not working.
        #! TODO Test this

        # def set_focus(widget: Widget) -> None:
        #     """Callback to set the focus."""
        #     try:
        #         widget.screen.set_focus(self, scroll_visible=False)
        #     except NoScreen:
        #         pass

        # self.app.call_later(set_focus, self)
        return self  
        
    def on_focus(self):

        self.log.debug("Triggered: on_focus in TextAreaReply")
        self.parent.parent.toot_widget.on_focus()

        # The parent.parent thing is messier than messages. But sometimes its less effort. ¯\_(ツ)_/¯
        # Especially when you need to do it a bunch of times in a row, like here.
        # This is probably a bad habit though.

    def on_blur(self):
        self.parent.parent.toot_widget.on_blur()

    def action_submit(self):
        self.post_message(self.Submit())

    async def action_cancel(self):
        await self.parent.parent.toot_widget.reply_to_toot()
        

class TootBox(Container):


    def __init__(self, toot_widget=None, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.toot_widget = cast(Widget, toot_widget)      # If toot_widget is NOT None, this is a reply
        self.hidden = False
        self.saved_text = None

    def compose(self):
        with Container(id="toot_box_container", classes="toot_box_container"):
            if self.toot_widget:        # different Bindings if its a reply vs a main toot
                yield TextAreaReply(id="toot_box_input", classes = "toot_box_input")
            else:
                yield TextAreaMain(id="toot_box_input", classes = "toot_box_input")
        with Container(classes="toot_box_container"):
            with Horizontal(classes="reply_footer"):
                yield SimpleButton("Post Toot", id="toot_box_reply", classes="toot_box_button")  # always present in center
                if not self.toot_widget:      # no attached widget means this is the main toot box
                    yield SimpleButton("Search mode", id="toot_box_search", classes="toot_box_button left")
                    yield SimpleButton("Hide", id="toot_box_hide", classes="toot_box_button right")
                else:                  # attached widget means this is a reply, and we want a cancel button
                    yield SimpleButton("Cancel", id="toot_box_cancel", classes="toot_box_button")

    def on_mount(self):

        self.query_one("#toot_box_reply").can_focus = False
        if not self.toot_widget:
            self.query_one("#toot_box_search").can_focus = False
            self.query_one("#toot_box_hide").can_focus = False
        else:
            self.query_one("#toot_box_cancel").can_focus = False

        self.toot_box_container = self.query_one("#toot_box_container")

        if self.toot_widget:
            self.input_box = cast(TextAreaReply, self.query_one("#toot_box_input"))
            self.focus_tootbox()
        else:
            self.input_box = cast(TextAreaMain,  self.query_one("#toot_box_input"))


    ###~ UTILITY METHODS ~###

    def focus_tootbox(self) -> None:
        self.input_box.focus()
        self.input_box.action_cursor_line_end()

    def set_text(self, text: str) -> None:
        """Set text in toot box and focus.
        Called by:
        - tootscreens.TootOptionsOtherUser.mention_thingy 
        - toot.TootWidget._delete_toot (when redrafting) """
        self.input_box.text = text
        self.focus_tootbox()

    @on(SimpleButton.Pressed, selector="#toot_box_cancel")
    def cancel_reply(self):
        self.toot_widget.replybox_on = False
        self.remove()

    @on(TextAreaMain.Search)
    @on(SimpleButton.Pressed, selector="#toot_box_search")
    def search_mode(self):
        self.log.debug("Search key pressed")
        self.app.push_screen(NotImplementedScreen('Search mode', classes="modal_screen"))  # TODO Implement


    @on(TextAreaMain.Hide)
    async def key_hide(self):
        self.log.debug("Hide key pressed")
        await self.hide_tootbox()

    @on(TextAreaReply.Submit)
    @on(TextAreaMain.Submit)
    async def key_submit(self):
        self.log.debug("Submit key pressed")
        await self.post_toot()


    ###~ MAIN LOGIC ~###

    @on(SimpleButton.Pressed, selector="#toot_box_reply")
    async def post_toot(self) -> None:

        if self.app.safe_mode:
            self.post_message(SuperNotify("Please disable safe mode to post."))
            return
        if not self.app.mastodon:
            self.post_message(SuperNotify("You're not logged in."))
            return
        if self.hidden:     # this should never happen.
            self.log.error("Tried to post toot with box hidden")
            return
        if self.input_box.text == "":
            self.post_message(SuperNotify("You can't post an empty toot."))
            return
        
        posted_toot = None
        idempotency_key = str(uuid.uuid4())   # good practice. prevents double posting

        if self.toot_widget:          # attached toot widget means this is a reply
            temp_id = self.toot_widget.boosted_id if self.toot_widget.reblog else self.toot_widget.toot_id
            with self.app.capture_exceptions():
                posted_toot = await self.app.mastodon.status_reply(
                    status = self.input_box.text,
                    to_status = self.toot_widget.json,
                    in_reply_to_id = temp_id,
                    idempotency_key = idempotency_key
                )           
        else:               # no attached widget - this is a new toot
            with self.app.capture_exceptions():                          
                posted_toot = await self.app.mastodon.status_post(
                    status = self.input_box.text,
                    idempotency_key = idempotency_key
                )
            
        if posted_toot:
            self.post_message(SuperNotify("Toot posted successfully."))
            self.input_box.text = ""
            if self.toot_widget:    # if this is a reply
                await self.toot_widget.query_one("#reply_box").remove()
                self.toot_widget.replybox_on = False
                self.toot_widget.refresh_toot()

    def set_memory_text(self):
        self.set_text(self.saved_text)
        self.saved_text = None

    @on(SimpleButton.Pressed, selector="#toot_box_hide")
    async def hide_tootbox(self) -> None:
        if self.hidden:
            self.hidden = False
            self.toot_box_container.mount(TextAreaMain(id="toot_box_input", classes="toot_box_input"))
            self.query_one("#toot_box_hide").update("Hide")
            self.query_one("#toot_box_hide").can_focus = False
            self.query_one("#toot_box_reply").visible = True
            self.query_one("#toot_box_search").visible = True
            self.set_styles("height: auto;")
            self.input_box = self.query_one("#toot_box_input")

            # this saves what the user was typing if they hide the box and then show it again
            if self.saved_text:
                self.set_timer(self.app.text_insert_time, self.set_memory_text)
        else:
            if self.input_box.text != "":           # if the input is not empty, save it
                self.log(f"Saving text: {self.input_box.text}")
                self.saved_text = self.input_box.text
            self.hidden = True
            self.query_one("#toot_box_input").remove()
            self.query_one("#toot_box_hide").update("Show")
            self.query_one("#toot_box_hide").can_focus = True
            self.query_one("#toot_box_reply").visible = False
            self.query_one("#toot_box_search").visible = False
            self.set_styles("height: 1;")



# Standard library imports
from __future__ import annotations
from typing import TYPE_CHECKING, Callable #, cast

if TYPE_CHECKING:
    from textual.app import ComposeResult
    from textual.widget import Widget

# Third party imports
from textual_pyfiglet import FigletWidget
from rich.text import Text
# from mastodon import Mastodon

# Textual imports
from textual import on, work
from textual.message import Message
from textual.worker import Worker, WorkerState
from textual.containers import Horizontal, Container
from textual.widget import Widget
from textual.widgets import Static

# TextualDon imports
from textualdon.widgets import (
    HashtagWidget,
    NewsWidget,
    TimelineSelector,
    ProfileWidget
)
from textualdon.settings import Settings, DevSettings
from textualdon.toot import TootWidget
from textualdon.oauth import OAuthWidget
from textualdon.widgets import WelcomeWidget
from textualdon.simplebutton import SimpleButton
from textualdon.messages import UpdateBannerMessage, SuperNotify, SwitchMainContent
from textualdon.screens import NotImplementedScreen


class PageHeader(Horizontal):

    class RefreshPage(Message):
        """This message is sent when the refresh button is pressed in the PageHeader widget."""
        def __init__(self) -> None:
            super().__init__()

    def __init__(self, page: str, refresh_visible: bool = True, **kwargs) -> None:
        super().__init__(**kwargs)
        self.page = page
        self.refresh_visible = refresh_visible

    def compose(self):
        yield FigletWidget(self.page, justify="left")
        yield SimpleButton("Refresh", id="refresh_button", classes="header_button")

    def on_mount (self):
        self.query_one("#refresh_button").can_focus = False
        if not self.refresh_visible:
            self.query_one("#refresh_button").visible = False

    @on(SimpleButton.Pressed, "#refresh_button")
    async def start_refresh_page(self) -> None:
        self.post_message(self.RefreshPage())   

class Page(Container):

    limit = 10
    refresh_allowed = True      # most are allowed. Set to false on Login and About pages
    populated = False

    @on(PageHeader.RefreshPage)
    async def start_refresh_page(self):

        if not self.refresh_allowed:
            self.notify("This page cannot be refreshed.")
            return

        if self.app.safe_mode:
            self.post_message(SuperNotify("Please disable safe mode to refresh."))
            return
        
        if not self.app.mastodon:
            self.post_message(SuperNotify("You are not logged in."))
            return
        
        self.log(Text(f"{self.app.breaker_figlet}", style="green"))

        self.log.debug(f"Refreshing {self.timeline} page")
        self.post_message(UpdateBannerMessage(f"Refreshing {self.timeline} page"))

        with self.app.capture_exceptions():
            self.refresh_page()

    async def _refresh_page(
        self,
        widget_type: Widget,
        method: Callable,
        *args,
        **kwargs
    ):
        """*args and **kwargs are passed to the method that is called on the Mastodon object."""

        css_dict = {
            TootWidget: "toot",
            HashtagWidget: "hashtag",
            NewsWidget: "news",
            ProfileWidget: "profile",
        }
        css_class = css_dict[widget_type]

        method_obj = getattr(self.app.mastodon, method)
        json_response = None
        with self.app.capture_exceptions():
            json_response = await method_obj(*args, **kwargs)
        if json_response:
        
            children = list(self.query_children().results())
            self.log.debug(f"children: {children}")
            to_remove = []
            for child in children:
                if isinstance(child, PageHeader) or isinstance(child, TimelineSelector):
                    pass                 # dont want to remove the headers
                else:
                    to_remove.append(child)
            await self.remove_children(to_remove)  

            if not self.app.error:
                widgets = [
                    widget_type(name=f"{widget_type}_{index}", json=json, classes=f"page_box {css_class}") 
                    for index, json in enumerate(json_response)
                ]
                await self.mount_all(widgets)

    @on(Worker.StateChanged)
    def worker_state_changed(self, event: Worker.StateChanged) -> None:
        
        if event.state == WorkerState.SUCCESS:
            self.log(Text(f"Worker {event.worker.name} completed successfully", style="green"))
        elif event.state == WorkerState.ERROR:
            self.log.error(Text(f"Worker {event.worker.name} encountered an error", style="red"))
        elif event.state == WorkerState.CANCELLED:
            self.log(Text(f"Worker {event.worker.name} was cancelled", style="yellow"))


class LoginPage(Container):

    refresh_allowed = False

    def compose(self) -> ComposeResult:
        
        yield PageHeader("login & settings", refresh_visible=self.refresh_allowed)
        yield WelcomeWidget(classes="page_box message")
        yield OAuthWidget(id="oauth_widget")                          # these both have IDs so they can be queried
        yield Settings(id="settings_widget", classes="settings")           # by the main app.
        with Container(classes="page_box footer"):
            yield SimpleButton("About TextualDon", id="about_button")

    @on(SimpleButton.Pressed, selector="#about_button")
    def show_about_page(self) -> None:
        self.post_message(SwitchMainContent("about"))


class HomePage(Page):

    timeline = "home"

    def compose(self) -> ComposeResult:
        yield PageHeader(self.timeline)

    @work
    async def refresh_page(self):

        await self._refresh_page(
            widget_type = TootWidget,
            method = "timeline",
            limit = self.limit,
            timeline = self.timeline
        )

class NotificationsPage(Page):

    timeline = "notifications"
    refresh_allowed = False
    
    def compose(self) -> ComposeResult:
        yield PageHeader(self.timeline)
        yield Static("Notifications page is a work in progress", classes="page_box wip")

    @work
    async def refresh_page(self):

        return    #! WIP
    
        json_response = await self.app.mastodon.notifications()


class ExplorePage(Page):

    page = "explore"
    timeline = "explore_posts"

    timeline_list = [
        ('Posts', 'explore_posts'),         # 0
        ('Hashtags', 'explore_hashtags'),   # 1
        ('People', 'explore_people'),       # 2
        ('News', 'explore_news')            # 3
    ]
    
    def compose(self) -> ComposeResult:
        yield PageHeader(self.page)
        yield TimelineSelector(self.timeline_list)

    @on(TimelineSelector.ChangeTimeline)
    def change_timeline(self, event: TimelineSelector.ChangeTimeline) -> None:
        self.timeline = event.timeline
        self.log.debug(f"self.timeline is set to: {self.timeline}")       

    @work
    async def refresh_page(self):

        if self.timeline == "explore_hashtags":
            await self._refresh_page(HashtagWidget, "trending_tags", limit=self.limit)

        if self.timeline == "explore_news":
            await self._refresh_page(NewsWidget, "trending_links", limit=self.limit)

        if self.timeline == "explore_posts":
            await self._refresh_page(TootWidget, "trending_statuses", limit=self.limit)

        if self.timeline == "explore_people":
            self.app.push_screen(NotImplementedScreen("More pages"))


class LiveFeeds(Page):

    page = "livefeeds"
    timeline = "local"

    timeline_list = [
        ('This Server', 'local'),         # 0
        ('Other Servers', 'public'),      # 1
    ]

    def compose(self) -> ComposeResult:
        yield PageHeader(self.page)
        yield TimelineSelector(self.timeline_list)

    @on(TimelineSelector.ChangeTimeline)
    def change_timeline(self, event: TimelineSelector.ChangeTimeline) -> None:
        self.timeline = event.timeline
        self.log.debug(f"self.timeline is set to: {self.timeline}")

    @work
    async def refresh_page(self):

        await self._refresh_page(
            widget_type = TootWidget,
            method = "timeline",
            limit = self.limit,
            timeline = self.timeline
        )


class PrivateMentionsPage(Page):

    timeline = "private mentions"
    refresh_allowed = False
    
    def compose(self) -> ComposeResult:
        yield PageHeader(self.timeline)
        yield Static("Private Mentions page is a work in progress", classes="page_box wip")

    @work
    async def refresh_page(self):

        return    #! WIP

        # json_response = await self.app.mastodon.notifications(mentions_only=True)    


class BookmarksPage(Page):

    timeline = "bookmarks"
    
    def compose(self) -> ComposeResult:
        yield PageHeader(self.timeline)

    @work
    async def refresh_page(self):

        await self._refresh_page(
            widget_type = TootWidget,
            method = "bookmarks",
        )


class FavoritesPage(Page):

    timeline = "favorites"
    
    def compose(self) -> ComposeResult:
        yield PageHeader(self.timeline)

    @work
    async def refresh_page(self):

        await self._refresh_page(
            widget_type = TootWidget,
            method = "favourites",
        )


class ListsPage(Page):

    timeline = "lists"
    refresh_allowed = False
    
    def compose(self) -> ComposeResult:
        yield PageHeader(self.timeline)
        yield Static("Lists page is a work in progress", classes="page_box wip")

    @work
    async def refresh_page(self):

        return    #! WIP

        # json_response = await self.app.mastodon.lists()       


class TootPage(Page):

    timeline = "toot"            
    main_toot_id: int | None = None

    def compose(self) -> ComposeResult:
        yield PageHeader(self.timeline)

    @work
    async def refresh_page(self):
        """The logic here is too compex to re-use the _refresh_page method."""

        json_single, json_response = None, None
        with self.app.capture_exceptions():
            json_single   = await self.app.mastodon.status(self.main_toot_id)
            json_response = await self.app.mastodon.status_context(self.main_toot_id)  
        if json_single and json_response:

            await self.remove_children()

            # first mount the main toot
            self.mount(TootWidget(name=f"toot_{self.main_toot_id}", json=json_single, classes="page_box toot"))     

            # ancestors = json_response["ancestors"]       #! TODO not implemented yet
            descendants = json_response["descendants"]

            # mount 'Replies' label
            self.mount(FigletWidget('Replies', font="small", classes="page_box figlet"))

            widgets = [
                TootWidget(name=f"toot_{index}", json=json, classes="page_box toot") 
                for index, json in enumerate(descendants)
            ]
            
            await self.mount_all(widgets)

class UserProfilePage(Page):

    timeline = "user profile" 
    account:  dict | None = None
    relation: dict | None = None     

    def compose(self) -> ComposeResult:
        yield PageHeader(self.timeline)
        yield Static("User Profile page is a work in progress", classes="page_box wip")

    def update_user(self, account_dict: dict, relation_dict: dict) -> None:
        self.account = account_dict
        self.relation = relation_dict

    @work
    async def refresh_page(self):

        self.log.debug("Refreshing user profile page")

        await self.remove_children()

        # This page does not make any API calls itself. Pure dependency injection.

        self.mount(
            ProfileWidget(account_dict=self.account, relation_dict=self.relation, classes="page_box toot")
        )

        # TODO get user's toots and display them here


class TooSmallPage(Container):

    refresh_allowed = False

    def compose(self) -> ComposeResult:
        yield FigletWidget("too small", justify="left", classes="page_header figlet")
        yield Static("The window is too small to display the app. Please resize the window.", classes="page_box small")


class DevelopmentPage(Container):

    refresh_allowed = False

    def compose(self) -> ComposeResult:

        yield PageHeader("Development Settings", refresh_visible=self.refresh_allowed)
        yield DevSettings(classes="settings dev")


class AboutPage(Container):

    refresh_allowed = False
    git_repo = "http://www.github.com/edward-jazzhands/textualdon"

    about_text = f"""TextualDon - by Edward Jazzhands © Copyright 2024 \n
TextualDon is a Mastodon client built with the Textual framework for Python. \n
It is a work in progress and is not yet feature complete. \n
The code is available on GitHub at {git_repo} \n"""
    
    refresh_allowed = False

    # TODO add link buttons to this page, and more info 

    def compose(self):
        yield PageHeader("about", refresh_visible=self.refresh)
        yield Static(self.about_text, classes="page_box")


# Standard library imports
from __future__ import annotations
from typing import TYPE_CHECKING, cast
from queue import Queue
from http.server import HTTPServer, BaseHTTPRequestHandler
from urllib.parse import urlparse, parse_qs
import time
import re

if TYPE_CHECKING:
    from textual.app import ComposeResult 

# Third party imports
from mastodon import Mastodon
from rich.text import Text

# Textual imports
from textual import on, work
from textual.worker import Worker, WorkerCancelled, WorkerState
from textual.binding import Binding
from textual.containers import Container, Horizontal
from textual.widgets import Static
from textual.widgets import Input

# TextualDon imports
from textualdon.messages import (
    SuperNotify,
    LoginStatus,
    LoginComplete,
    CallbackSuccess,
)
from textualdon.simplebutton import SimpleButton
from textualdon.sql import SQLite    # only for casting
from textualdon.html_templates import html_success_msg, html_failure_msg
from textualdon.screens import CallbackScreen
from textualdon.savedusers import SavedUsersManager


class OAuthInput(Input):

    BINDINGS = [
        Binding("enter", "submit", "Submit", show=True)
    ]

    # @on(OAuthInput.Submitted)
    # async def submit_input(self, event: OAuthInput.Submitted) -> None:
    #     """This is here to allow submitting with the enter button. Custom binding on OAuthInput."""

    #     self.log.debug("Trigger function: submit_input in oauth.py")
    #     await self.oauth_flow()

class OAuthWidget(Container):


    client_id = None
    client_secret = None
    queue = Queue()
    callback_active = False

    def compose(self) -> ComposeResult:

        self.db = cast(SQLite, self.app.sqlite)

        self.callback_wait_time:    int = self.app.config.getint('MAIN', 'callback_wait_time')
        self.redirect_uri_template: str = self.app.config.get('MAIN', 'redirect_uri')

        with Horizontal(id="login_form"):
            yield OAuthInput(placeholder="mastodon.social", id="instance_url")
            yield SimpleButton("Connect", id="connect_button")
        yield SavedUsersManager(id="saved_users_manager")
        yield Static("Enter an instance URL or choose from saved logins.", id="login_status")

    async def on_mount(self) -> None:

        self.saved_users_manager = cast(SavedUsersManager, self.query_one(SavedUsersManager))
        self.login_status = cast(Static, self.query_one("#login_status"))
        self.login_input  = cast(Input, self.query_one("#instance_url"))
        self.query_one("#connect_button").can_focus = False

        # NOTE: app data only needs to be created once per Mastodon server instance.

        query = "SELECT * FROM app_data"
        app_data = self.db.fetchall(query)
        self.log.debug(f"Existing app data: {app_data}")

        await self.saved_users_manager.start_process()

    @on(SavedUsersManager.TriggerLogin)
    async def trigger_login(self, event: SavedUsersManager.TriggerLogin) -> None:
        """This is triggered by the SavedUsersManager widget when a user is selected."""

        self.log.debug("Trigger function: trigger_login in OAuthWidget")
        with self.app.capture_exceptions():
            await self.login_stage5(event.instance_url)

    # @on(OAuthInput.Submitted)
    # async def submit_input(self, event: OAuthInput.Submitted) -> None:

    #     self.log.debug("Trigger function: submit_input in oauth.py")
    #     await self.oauth_flow()

    ###~ Oauth Flow ~###

    @on(OAuthInput.Submitted) #! TEST THIS
    @on(SimpleButton.Pressed, '#connect_button')
    async def oauth_flow(self) -> None:

        if self.app.safe_mode:
            self.notify("Please disable safe mode first.", timeout=7)
            return

        stage1_passed = False
        stage2_passed = False
        stage3_passed = False
        stage4_passed = False

        stage1_result = None
        stage2_result = None
        stage3_result = None
        stage4_result = None

        instance_url:str = self.login_input.value
        self.login_input.value = ""   # clear the input field

        validated = True if re.match(r'\w+\.\w+', instance_url) else False
        if not validated:
            self.notify("Please enter a valid URL.")
            return

        port:str = self.db.fetchone("SELECT value FROM settings WHERE name = ?", ('callback_port',))[0]
        redirect_uri = self.redirect_uri_template.format(port=port)

        # check if app credentials are already saved for this instance
        app_data = self.db.fetchone("SELECT * FROM app_data WHERE instance_url = ?", (instance_url,))
        self.log.debug(f"app_data: \n{app_data}")
        
        if app_data:
            client_id = app_data[2]
            client_secret = app_data[3]
            self.log("App credentials found. Skipping stage 1.")
            stage1_passed = True

        else:
            self.log("Registering a new app wih the instance.")  

            with self.app.capture_exceptions():
                stage1: Worker = self.login_stage1(instance_url, redirect_uri)
                await stage1.wait()
                stage1_result: tuple[str, str] = stage1.result
            if self.app.error:
                return
            else:
                # Stage 1 result is a tuple of client ID:str and client secret:str
                self.log.debug(f"STAGE 1 RESULT: \n {stage1_result}")
                with self.app.capture_exceptions():
                    stage1_result[0].strip()
                    stage1_result[1].strip()
                if self.app.error:
                    return
                else:
                    stage1_passed = True
                    client_id, client_secret = stage1_result
                    self.db.insert_one(
                        'app_data',
                        ['instance_url', 'client_id', 'client_secret'],
                        [instance_url, client_id, client_secret]
                    )

        if stage1_passed is True:
            with self.app.capture_exceptions():                      
                stage2: Worker = self.login_stage2(client_id, client_secret, instance_url, redirect_uri)
                await stage2.wait()
                stage2_result: tuple[Mastodon, str] = stage2.result
            if self.app.error:
                return
            else:
                # Stage 2 result is a tuple of Mastodon object and auth URL
                self.log.debug(f"STAGE 2 RESULT: \n {stage2_result}")
                with self.app.capture_exceptions():
                    self.log(stage2_result[0].client_id)    # checks if we can access an attribute of the Mastodon object
                    self.log(stage2_result[1].strip())
                if self.app.error:
                    return
                else:
                    self.app.attach_mastodon(stage2_result[0])    #~ This is where the mastodon object becomes 
                    stage2_passed = True                          #~ globally accessible to the app.

        if stage2_passed is True:
            with self.app.capture_exceptions():    
                stage3_result = await self.login_stage3(stage2_result[1], int(port))
            if self.app.error:
                return
            else:
                # Stage 3 result is the temporary auth code from the callback server
                self.log.debug(f"STAGE 3 RESULT: \n {stage3_result}")
                with self.app.capture_exceptions():
                    stage3_result.strip()
                if self.app.error:
                    return
                else:
                    stage3_passed = True                 

        if stage3_passed is True:   
            self.post_message(CallbackSuccess())                      
            with self.app.capture_exceptions():    
                stage4_result: str = await self.login_stage4(stage3_result, redirect_uri)
            if self.app.error:
                return
            else:
                # Stage 4 result is the access token
                self.log.debug(f"STAGE 4 RESULT: \n {stage4_result}")
                with self.app.capture_exceptions():
                    stage4_result.strip()
                if self.app.error:
                    return
                else:
                    stage4_passed = True                 

        if stage4_passed is True:
            with self.app.capture_exceptions():
                await self.login_stage5(instance_url, access_token=stage4_result)

    # Stages 1 and 2 are both in the group 'login', and exclusive=True.
    # it probably doesn't matter much but it might help prevent it being
    # run twice at the same time for some reason.

    @work(thread=True, exclusive=True, exit_on_error=False, group="login")
    def login_stage1(self, instance_url: str, redirect_uri: str) -> tuple[str, str]:
        """Stage 1: Registers the app with the Mastodon instance.   
        This should only be run once and it will save the app profile to a file.
        
        Returns:
            tuple[str, str]: The client ID and client secret."""

        self.log("Starting Worker: Login Stage 1")

        if instance_url is None:
            raise ValueError("Instance URL not provided.")  # this shouldn't ever happen

        try:
            client_id, client_secret = Mastodon.create_app(
                "TextualDon",
                redirect_uris = redirect_uri,
                api_base_url = instance_url,
            )
        except Exception as e:
            raise e
        
        client_id: str
        client_secret: str
        return client_id, client_secret         # Passed into stage 2

    @work(thread=True, exclusive=True, exit_on_error=False, group="login")
    def login_stage2(
            self,
            client_id: str,
            client_secret: str,
            instance_url: str,
            redirect_uri: str
        ) -> tuple[Mastodon, str]:
        """Stage 2: Submits app credentials to the Mastodon instance.   
        Requests an auth URL from the instance, opens the browser.  
        
        Returns:
            tuple[Mastodon, str]: The Mastodon object and the auth url."""
        
        self.log("Starting Worker: Login Stage 2")
        self.log.debug(f"Client ID: {client_id} | Client Secret: {client_secret}")

        if not client_id or not client_secret:
            raise ValueError("Client ID or Client Secret not found.")
    
        try:
            mastodon = Mastodon(              #~ Here we create the Mastodon object that will be used,
                client_id = client_id,          # But only if its the first time logging in to this server.
                client_secret= client_secret,       # Following times are handled in savedusers.py
                api_base_url = instance_url,
                version_check_mode='none',
                request_timeout=3
            )     
        except Exception as e:
            raise e                                       

        try:            #! TODO: I think this wants 'state' for idempotency
            auth_url = mastodon.auth_request_url(redirect_uris = redirect_uri)
        except Exception as e:
            raise e

        return mastodon, auth_url

    async def login_stage3(self, auth_url: str, port: int) -> str:

        self.log("Starting Worker: Login stage 3")

        await self.app.push_screen(CallbackScreen(auth_url, id="callback_screen", classes='modal_screen'))

        try:
            callback: Worker = self.run_callback_server(port)
            await callback.wait()
        except WorkerCancelled as e:
            raise e
        except Exception as e:
            raise e
        else:
            if isinstance(callback.result, str):
                return callback.result
            else:
                raise ValueError("Callback result is not a string.")
        finally:
            self.log(f"callback:  {callback}  |  callback.result:  {callback.result}")

    async def login_stage4(self, auth_code: str, redirect_uri: str) -> str:
        """Stage 4: Logs in with the auth code received from the callback server.
        
        Returns:
            str: The access token."""

        self.log("Starting Worker: Login Stage 4")
        if not auth_code:
            raise ValueError("No auth code received.")

        access_token = await self.app.mastodon.log_in(redirect_uri = f'{redirect_uri}', code=auth_code)

        if not isinstance(access_token, str):
            raise ConnectionError(f"Could not log in, recieved: {access_token}")

        return access_token

    async def login_stage5(self, instance_url: str, access_token: str | None = None) -> None:
        """Stage 5: Updates the UI and Database.

        THIS FUCTION IS CALLED BY:
        - OauthWidget.oauth_flow
        - OauthWidget.trigger_login

        Passing in an access token means its the first time logging in with that account.   
        
        Returns:
            str: The display name of the logged in user. """

        self.log("Starting Login Stage 5")


        app_cred = await self.app.mastodon.app_verify_credentials()
        account_cred = await self.app.mastodon.account_verify_credentials()

        if not isinstance(app_cred, dict) or not isinstance(account_cred, dict):
            raise ValueError("Could not verify credentials.")

        app_id = app_cred["id"]
        app_scopes = app_cred["scopes"]

        self.log(
            f"Mastodon object: {type(self.app.mastodon)} \n"
            f"Connected App ID: {app_id} | Scopes: {app_scopes}"
        )
        
        username = account_cred["username"]
        user_id = account_cred["id"]
        display_name = account_cred["display_name"]

        self.log(f"Logged in as {display_name} (@{username}) on {instance_url}")

        existing_user = self.db.fetchone("SELECT * FROM users WHERE id = ?", (user_id,))
        if existing_user:
            self.log.debug(f"User already exists in the database: {existing_user}")
        else:
            self.log.debug(f"User not detected in db: {existing_user}")
            columns = ["id", "username", "display_name", "instance_url", "access_token"]
            values = [user_id, username, display_name, instance_url, access_token]
            self.db.insert_one('users', columns, values)

        # NOTE: Technically, the user_id is not unique across the entire Mastodon network.
        # But since this is a client-side app, we can assume it is unique for our purposes.
        # The probability of a user having the same ID number on two instances is astronomically low.

        self.post_message(SuperNotify(f"Logged in as {display_name} (@{username})"))
        self.post_message(LoginStatus(
            statusbar=username,
            instance_url=instance_url,
            loginpage_message=f"Logged in as {display_name} (@{username}) on {instance_url}",
        ))

        self.log(f"Logged in successfully to {username}, user ID: {user_id}")

        self.app.logged_in_user_id = user_id
        self.db.update_column('settings', 'value', user_id, 'name', 'last_logged_in')

        # Refresh the display name every time we use a saved log in.
        if access_token is None:
            self.db.update_column('users', 'display_name', display_name, 'id', user_id)

        await self.saved_users_manager.start_process()  # refresh the saved users list
        self.app.instance_url = instance_url            # make accessible at app level

        self.log(Text(f"Login Successful. \n {self.app.breaker_figlet}", style="green"))
        self.app.post_message(LoginComplete()) 

    ###~ Callback Server ~###

    @work(thread=True, exclusive=True, exit_on_error=False, group='callback')
    async def run_callback_server(self, port: int) -> str:

        self.log("Starting Worker: callback server.")

        server_address = ('', port)
        OAuthCallbackHandler.queue = self.queue          # dependency inject the queue
        httpd = HTTPServer(server_address, OAuthCallbackHandler)
        httpd.timeout = 5                       # Set a low timeout to periodically exit handle_request
        start_time = time.time()

        self.log.debug(f"Callback server address: {httpd.server_address}")

        self.callback_active = True # this is a flag we can modify in another method to stop the loop

        # this will time out every 5 seconds and restart. It makes it more reliable/responsive.
        try:
            while (time.time() - start_time) < self.callback_wait_time and self.callback_active:
                httpd.handle_request()  # Handles a single request or timeout
                if self.queue.qsize() > 0:
                    self.log.debug("Queue size > 0; exiting callback server.")
                    break
                self.log.debug("No OAuth callback yet; retrying...")
        except Exception as e:
            raise e
        finally:
            httpd.server_close()
            self.callback_active = False
            self.log.debug("Callback server closed.")

        if self.queue.qsize() > 0:
            return self.queue.get()

    def cancel_callback(self) -> None:

        cancelled = self.workers.cancel_node(self.query().node)
        self.callback_active = False   # required to make the loop exit after the worker is cancelled

        self.log.debug(Text("Cancelling callback server. \n"
                        f"Cancelled workers: {cancelled}", style="yellow"))
        

    ###~ Worker Debug stuff ~###

    @on(Worker.StateChanged)
    def worker_state_changed(self, event: Worker.StateChanged) -> None:

        self.log.debug(Text(
                f"Worker state: {event.state}\n"
                f"Worker.name: {event.worker.name}",
                style="cyan"
        ))

        if event.state == WorkerState.SUCCESS:
            self.log(Text(f"Worker {event.worker.name} completed successfully", style="green"))
        elif event.state == WorkerState.ERROR:
            self.log.error(Text(f"Worker {event.worker.name} encountered an error", style="red"))
        elif event.state == WorkerState.CANCELLED:
            self.log(Text(f"Worker {event.worker.name} was cancelled", style="yellow"))


#* Will be run in the thread with run_callback_server:
class OAuthCallbackHandler(BaseHTTPRequestHandler):

    queue = None

    def do_GET(self):
        parsed_url = urlparse(self.path)       
        params = parse_qs(parsed_url.query)
        
        if 'code' in params:       
            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            self.queue.put(params['code'][0])
            
            message = html_success_msg
        else:
            self.send_response(400)
            self.end_headers()
            message = html_failure_msg

        self.wfile.write(message.encode('utf-8'))
















# Third party imports
from bs4 import BeautifulSoup

# Textual imports
from textual.widget import Widget


class BS4Parser(Widget):


    def parser(self, html_content: str) -> str:
        
        soup =  BeautifulSoup(html_content, 'html.parser')

        #? notes/bs4_cheatsheet.md

        markdown_list = []
        markdown_str = ""

        for element in soup.find_all('br'):
            element.replace_with("\n")

        for element in soup.find_all('span', attrs={'class': 'invisible'}):
            element.decompose()

        for element in soup.find_all('span', attrs={'class': 'ellipsis'}):
            if element.string:
                element.string = element.string + "..."

        for element in soup.find_all('p'):        
            markdown_list.append(element.get_text(strip=False))       
            # markdown += f"{element.get_text(strip=True)}"

        # self.log.debug(markdown_list)

        # I believe this has to be separate from the second enumerate, because
        # the list is being modified while it loops. So it changes the index of items.
        # It might be possible to combine them into one loop, but it was giving me issues.
        # so fuck it, they're two loops.
        for index, element in enumerate(markdown_list):
            if element == "":
                markdown_list.pop(index)

        for index, element in enumerate(markdown_list):
            if index != len(markdown_list) - 1:     # if not last element,
                markdown_str += f"{element}\n\n"    # add 2 newlines
            else:
                markdown_str += f"{element}"        # last element does not get newlines


        return markdown_str
    


# EXAMPLE:

"""
<p>                                                                                                                                             
    Rocket Report: ULA 'supremely confident' in Vulcan; Impulse Space rakes it in                                                                  
</p>                                                                                                                                            
<p>                                                                                                                                             
    "I'm pretty darn confident I'm going to have a good day on Friday."                                                                          ▅▅
</p>                                                                                                                                            
<p>                                                                                                                                             
<a href="https://arstechnica.com/space/2024/10/rocket-report-ula-supremely-confident-in-vulcan-impulse-space-rakes-it-in/?utm_brand=arstechnica&amp;utm_social-type=owned&amp;utm_source=mastodon&amp;utm_medium=social" 
rel="nofollow noopener noreferrer" target="_blank" translate="no">                                                                              
<span class="invisible">                                                                                                                      
    https://                                                                                                                                     
</span>                                                                                                                                       
<span class="ellipsis">                                                                                                                       
    arstechnica.com/space/2024/10/                                                                                                               
</span>                                                                                                                                       
<span class="invisible">                                                                                                                      
    rocket-report-ula-supremely-confident-in-vulcan-impulse-space-rakes-it-in/?utm_brand=arstechnica&amp;utm_social-type=owned&amp;utm_source=mastodon&amp;utm_medium=social                                                                                 
</span>                                                                                                                                       
</a>                                                                                                                                           
</p>
"""


# Standard Library Imports
from typing import List, Dict, Tuple, cast
from sqlite3 import DatabaseError
from pathlib import Path
from datetime import datetime

# Third party imports
from rich.console import Console
from rich.traceback import Traceback
from mastodon import (
    MastodonError, 
    MastodonNetworkError, 
    MastodonUnauthorizedError,
    MastodonNotFoundError
)
from clipman.exceptions import ClipmanBaseException
from pyperclip import PyperclipException

# Textual Imports
from textual import on
from textual.screen import ModalScreen, Screen
from textual.containers import Container, Horizontal, VerticalScroll
from textual.widgets import Button, Label, TextArea
from textual.dom import DOMNode
from textual.binding import Binding
from textual.worker import WorkerError, WorkerFailed, WorkerCancelled
from textual.errors import TextualError
from textual.css.query import NoMatches

# Textualdon Imports
from textualdon.messages import UpdateBannerMessage, SuperNotify
from textualdon.simplebutton import SimpleButton


class SafeModeError(Exception):
    """Raised when the app is in safe mode and a restricted action is attempted."""
    pass


class ErrorHandler(DOMNode):

    def __init__(self, data_dir: Path, **kwargs):
        """Initialize the Error Handler.

        Args:
            data_dir: The directory where the app's data is stored.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes for the widget. """

        super().__init__(**kwargs)
        self.data_dir = data_dir
        self.logs_dir = self.data_dir / "logs"
        self.logfile = None     # Will be set in the handle_exception method.
        self.loghtml = None     # Will be set in the handle_exception method.

        self.logs_dir.mkdir(parents=True, exist_ok=True)
        self.super_error_dict = {}
        self.error_number = 0

    def reset_stored_errors(self):
        """Triggered by app.disable_safe_mode()."""
        self.super_error_dict = {}
        self.error_number = 0

    async def recursive_cause(self, e, error_list) -> List:

        error_list.append(e)

        if hasattr(e, '__notes__') and e.__notes__:
            error_list.append(e.__notes__)

        if hasattr(e, '__cause__') and e.__cause__:    # Check for explicit chaining
            await self.recursive_cause(e.__cause__, error_list)
        elif hasattr(e, '__context') and e.__context__:  # Check for implicit chaining
            await self.recursive_cause(e.__context__, error_list)

        return error_list

    async def handle_exception(self, e: Exception):

        self.log.debug(f"Exception traceback in ErrorHandler: {e.__traceback__}")
        logfile = self.logs_dir / f"error_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S_%f')}.txt"
        loghtml = self.logs_dir / f"error_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S_%f')}.html"
        self.logfiles = [logfile, loghtml]

        if isinstance(e, SafeModeError):
            self.app.notify("Please disable safe mode to perform this action.", timeout=7)
            return

        try:
            cause = e.__cause__                         # I'm honestly not sure if any of this
            cause_type = type(cause).__name__           # section here is really necessary.
        except: # noqa: E722
            cause = None
            cause_type = None
        try:
            context = e.__context__
            context_type = type(context).__name__
        except: # noqa: E722
            context = None
            context_type = None

        self.log.error(
            f"Exception type: {type(e)}\n"
            f"Cause: {cause_type}: {cause}\n"
            f"Context: {context_type}: {context}\n"
            f"args: {e.args}"
        )

        error_list = []
        try:
            error_msg_list = await self.recursive_cause(e, error_list)
        except: # noqa: E722
            self.log.error("Recursive error cause failed.")
            error_msg_list = [e]
        self.log.debug(error_msg_list)

        if isinstance(e, WorkerFailed):
            #* If a worker failed, its either going to be the Mastodon Proxy class's "APIRunner",
            #* or it'll be some essential part of the app.
            if isinstance(e.error, MastodonError):
                await self.handle_mastodon_error(e.error, error_msg_list)
                return
            
        elif isinstance(e, WorkerCancelled):
            self.log(f"Worker was cancelled: {e}")
            return

        elif isinstance(e, ClipmanBaseException):
            self.app.post_message(UpdateBannerMessage(f"Clipman +failed: {error_msg_list[-1]}"))
            self.app.notify(f"Clipman failed: {error_msg_list[-1]}")
            return
        
        elif isinstance(e, PyperclipException) or "clip.exe" in str(e).lower():
            self.app.post_message(UpdateBannerMessage(f"Pyperclip +failed: {error_msg_list[-1]}"))
            self.app.notify(f"Pyperclip failed: {error_msg_list[-1]}")
            return
        
        #* Then if we cant handle it in some above manner, save it to the super error dict,
        #* and increase the error_number by 1. We only make an addition and increment the counter
        #* if we can't handle the error in some other way.
        #* Errors are stored until the user disables safe mode.

        self.error_number += 1
        self.super_error_dict[self.error_number] = (error_msg_list, self.logfiles)

        #* Any exceptions that we can't handle above get written to a log file, then the user
        #* is shown the more serious App Error Screen, which will suggest they restart the program
        #* and consider posting the error to the issue tracker on Github.

        self.log(f"Current screen stack: {self.app.screen_stack}")
        for screen in self.app.screen_stack:        
            if screen.name == "error_screen":       # prevents having more than 1 error screen.
                screen.update_errors(self.super_error_dict)
                return
                #* we only care about logging the traceback once, for the first error.
                #* all other errors are caused by the first error so we don't need to log them.

        try:
            self.app.enter_safe_mode()
        except Exception as safe_error:
            self.log.error(f"Failed to enter safe mode: {safe_error}")

        try:      #~ Here writes error to log file
            await self.record_log_files(e, logfile, loghtml)
        except Exception as log_error:
            self.log.error(
                "Error writing error report to file. \n"
                f"Error: {log_error}"
            )
        else:
            self.log("Error report written to file.")

        if isinstance(e, MastodonError):
            component = 'mastodon.py'
        elif isinstance(e, DatabaseError):
            component = 'database'
        elif isinstance(e, TextualError) or isinstance(e, NoMatches):
            component = 'textual'
        elif isinstance(e, WorkerError):
            component = 'worker'
        else:
            component = 'python'

        try:
            await self.app.push_screen(ErrorScreen(
                    exception=e,
                    error_msg_list=error_msg_list,
                    component=component,
                    classes="modal_screen",
                    name="error_screen"
                ),
                callback=self.push_report_screen
            )
        except: # noqa: E722
            self.log.error("ERROR pushing error screen.")
            raise e         # if this fails, we want to see the error in the terminal.
        else:
            self.log("Error screen pushed.")

    async def handle_mastodon_error(self, e: Exception, error_msg_list: List):

        #* These will mostly be network or permission related errors. We can just notify the user
        #* and continue the program.

        self.log.debug(f"Mastodon Error type: {type(e)}")
        e_str = str(e)
        e_str_list:list = e_str.replace("'", "").split(", ")
        # deepest_e_str = e_str_list[-1]
        deepest_msg = error_msg_list[-1]

        if isinstance(e, MastodonNetworkError):
            self.app.notify("Network Error: URL doesn't exist, or your internet is down.", timeout=7)
            self.app.post_message(UpdateBannerMessage(f"Mastodon server says: {deepest_msg}"))
            return
        
        if isinstance(e, MastodonUnauthorizedError):
            self.app.notify(f"Mastodon server says: {deepest_msg}", timeout=7)
            self.app.post_message(UpdateBannerMessage(f"Mastodon server says: {deepest_msg}"))
            return

        if isinstance(e, MastodonNotFoundError):
            self.app.post_message(SuperNotify("404: Requested record not found."))
            return

        self.app.post_message(SuperNotify(f"Mastodon Error: {deepest_msg}"))
        self.log.debug(f"Mastodon Error: {e_str_list}")
        

    async def push_report_screen(self, result):
        """Callback function from the ErrorScreen."""

        try:
            await self.app.push_screen(ReportScreen(
                exception=result[0],
                super_error_dict=self.super_error_dict,
                component=result[1],
                classes="fullscreen"
            ))
        except: # noqa: E722
            self.log.error("Error pushing report screen.")
            raise self.e    # safety in case this process fails.
        else:
            self.log("Report screen pushed successfully.")

    async def record_log_files(self, e: Exception, logfile: Path, loghtml: Path):
        """Record the log files to the app's logfiles list."""

        # TODO Check amount of log files and delete oldest over a certain amount (10?)

        with open(logfile, "w") as report_file:
            console = Console(file=report_file, record=True, width=100)
            console.print(f"TextualDon Error Report: {datetime.now()}\n")
            # console.print_exception(show_locals=False)
            traceback = Traceback.from_exception(
                type(e),
                e,
                e.__traceback__,
                show_locals=False
            )
            console.print(traceback)            
            console.save_html(loghtml)

    async def delete_logs(self):
        """Delete all log files."""
        try:
            logs = list(self.logs_dir.iterdir())
            if not logs:
                self.app.post_message(SuperNotify("No log files to delete."))
                return
            for log in logs:
                log.unlink()
        except Exception as e:
            self.log.error(f"Error deleting logs: {e}")
            self.app.post_message(SuperNotify("Error deleting log files"))
        else:
            self.log("Logs deleted.")
            self.app.post_message(SuperNotify("Log files deleted."))

# NOTE: I really would have preferred to combine a bunch of commonality of the two screens
# here into one parent class. However, one is a regular Screen and the other is a ModalScreen.
# I'm sure there's a way to make that work but I didn't feel like digging into it right now.

class ErrorScreen(ModalScreen):

    BINDINGS = [
        Binding('q', 'quit', 'Quit'),
        Binding('i', 'ignore', 'Ignore'),
        Binding("up,left", "focus_previous", description="Focus the previous button."),
        Binding("down,right", "focus_next", description="Focus the next button."),
    ]

    controls = "Arrow keys, Tab/Shift+Tab: navigate | Enter: select | q: quit | i: ignore"   
        
    generic_warning = """TextualDon has encountered an unexpected error. 
Its recommended to quit the app immediately. If you have something you were working \
on, you can attempt to hit ignore and save your work. But be warned, there is a chance \
the app will freeze or crash as soon as you press the ignore button.

For full error information and to file a report, please press 'Read and Report' \n"""

    db_warning = """The database has encountered an error.
Ignoring this warning is disabled. The app will not work without the database.

For full error information and to file a report, please press 'Read and Report'  \n"""

    mastodon_warning = """The Mastodon connection returned an unexpected error. \n """


    def __init__(
        self, 
        exception: Exception, 
        error_msg_list: List, 
        component: str,
        **kwargs
    ):
        """Args:
            exception: The exception that was raised.
            error_msg_list: A list of error messages.
            logfiles: A list of log files to display to the user.
            component: The component that raised the error.
            name: The name of the screen.
            id: The ID of the screen in the DOM.
            classes: The CSS classes for the screen. """

        super().__init__(**kwargs)

        self.log(
            "Error screen initialized with exception: "
            f"{exception} and error messages: \n {error_msg_list}\n"
            f"component: {component}"
        )

        self.exception = exception
        self.error_msg_list = error_msg_list
        self.component = component
        self.super_error_count = 1

        if component == 'mastodon':
            self.warning = self.mastodon_warning
            self.banner = "[yellow]Mastodon API Error[/yellow]"
        elif component == 'database':
            self.warning = self.db_warning
            self.banner = "[red]! DATABASE ERROR ![/red]"
        else:
            self.warning = self.generic_warning
            self.banner = "[red blink]! ERROR ![/red blink]"

    def compose(self):

        with Container(classes='screen_container ultrawide'):
            yield Label(self.banner, classes='screen_label')
            yield Label(self.warning, classes='screen_label')
            yield SimpleButton("Show Error", id='show_error', classes='screen_button')
            with VerticalScroll(id="error_scroll"):
                yield Label("Size of error stack: 1", id="error_counter", classes='screen_label left')
                yield Label(f"Component: {self.component}", classes='screen_label left')
                yield Label(f'{type(self.exception).__name__}: {self.error_msg_list[-1]}', classes='screen_label left')
            yield SimpleButton("Read & Report", id='read_report', classes='screen_button')
            with Horizontal(classes='screen_buttonbar'):
                yield Button('Ignore', id='ignore', classes='screen_button large')
                yield Button('Quit', id='quit', classes='screen_button large')
        with Container(classes='screen_container wide help'):
            yield Label(self.controls, classes='screen_label')

    def on_mount(self):

        if self.component == 'database':
            self.query_one('#ignore').disabled = True

        self.query_one("#error_scroll").display = False

    @on(Button.Pressed, selector='#ignore')
    def ignore(self):
        if self.component == 'database':
            self.app.notify("You can't ignore this")
        else:
            self.app.pop_screen()
    
    @on(Button.Pressed, selector='#quit')
    def quitme(self):
        self.app.exit()

    @on(SimpleButton.Pressed, selector='#read_report')
    def read_report(self):
        result = (self.exception, self.component)
        self.dismiss(result)

    @on(SimpleButton.Pressed, selector='#show_error')
    def show_error(self):
        self.query_one("#error_scroll").display = not self.query_one("#error_scroll").display

    def update_errors(self, super_error_dict: Dict):
        self.query_one("#error_counter").update(f"Size of error stack: {len(super_error_dict)}")

    ###~ Keybindings ~###

    def action_focus_previous(self):
        self.focus_previous()
    
    def action_focus_next(self):
        self.focus_next()

    def action_quit(self):
        self.quitme()

    def action_ignore(self):
        self.ignore()
    

class ReportScreen(Screen):

    BINDINGS = [
        Binding('q', 'quit', 'Quit'),
        Binding('i', 'ignore', 'Ignore'),
        Binding("up,left", "focus_previous", description="Focus the previous button."),
        Binding("down,right", "focus_next", description="Focus the next button."),
    ]

    controls = "Arrow keys, Tab/Shift+Tab: navigate | Enter: select | q: quit | i: ignore"   
    gitrepo_path = "https://github.com/edward-jazzhands/textualdon/issues"

    def __init__(
        self, 
        exception: Exception, 
        super_error_dict: Dict,
        component: str,
        **kwargs
    ):
        """
        Args:
            exception: The exception that was raised.
            error_msg_list: A list of error messages.
            logfiles: A list of log files to display to the user.
            component: The component that raised the error.
            name: The name of the screen.
            id: The ID of the screen in the DOM.
            classes: The CSS classes for the screen. """

        super().__init__(**kwargs)
        self.exception:   Exception = exception
        self.super_error_dict: Dict = super_error_dict
        self.component:         str = component

    def compose(self):

        with VerticalScroll(classes='fullscreen'):
            with Container(classes='fullscreen container bordered_red'):
                yield Label("Exceptions are stored here until safe mode is disabled.", classes='screen_label')
                yield TextArea(id="error_box", read_only=True)
                yield SimpleButton("Copy above text to clipboard", id='copy_errors', classes='screen_button')
                yield SimpleButton("Copy log file path to clipboard", id='copy_path', classes='screen_button')
                yield SimpleButton(
                    "Open Rich traceback in your browser (works offline)", id='open_browser', classes='screen_button'
                )
                yield SimpleButton("Open Github issues page", id='open_github', classes='screen_button')
                yield Label(
                    "The logs folder contains both a plain txt version and an html version. \n"
                    "Log file location:", classes='screen_label'
                )
                yield TextArea(id="link_box", read_only=True, classes="link_box")
                yield TextArea(id="github_link", read_only=True, classes="link_box")
                with Horizontal(classes='screen_buttonbar'):
                    yield Button('Ignore & return', id='ignore', classes='screen_button large')
                    yield Button('Quit', id='quit', classes='screen_button large')
            with Container(classes='screen_container wide help'):
                yield Label(self.controls, classes='screen_label')

    def on_mount(self):

        self.error_box = cast(TextArea, self.query_one('#error_box'))
        self.error_box = self.query_one('#error_box')

        if self.component == 'database':
            self.query_one('#ignore').disabled = True

        self.set_timer(self.app.text_insert_time, self.set_text)

    def set_text(self):

        self.error_box.insert(
            f"Component that failed: {self.component}\n\n",
            maintain_selection_offset=False
        )

        # key is 1-based index. Each value in dict is tuple: (error_msg_list: List, self.logfiles: List)

        for key, value in self.super_error_dict.items():
            if key == 1:
                self.error_box.insert("Causing error set:\n\n", maintain_selection_offset=False)
            else:
                self.error_box.insert(f"Error set {key}:\n\n", maintain_selection_offset=False)

            for error in value[0]:
                error_type: str = type(error).__name__
                if error_type == 'list':
                    error_type = 'Notes'
                    self.error_box.insert("Notes:\n", maintain_selection_offset=False)
                    for note in error:
                        self.error_box.insert(
                            f"{note}\n",
                            maintain_selection_offset=False
                        )
                else:
                    self.error_box.insert(
                        f"{error_type}: {error}\n\n",
                        maintain_selection_offset=False
                    )

        # key 1 = first error, 2 = second error, etc
        # Each value is a tuple of the error message list and the log files list.
        # [1][0][0] = first error, error message list, first error message
        # [1][1][0] = first error, log files list, first log file path

        self.query_one('#link_box').insert(str(self.super_error_dict[1][1][0]))
        self.query_one('#github_link').insert(self.gitrepo_path)

    @on(Button.Pressed, selector='#ignore')
    def ignore(self):
        if self.component == 'database':
            self.app.notify("You can't ignore this")
        else:
            self.app.pop_screen()

    @on(Button.Pressed, selector='#quit')
    def quitme(self):
        self.app.exit()

    @on(SimpleButton.Pressed, selector='#copy_errors')
    def copy_errors(self):
        self.app.copy_to_clipboard(str(self.error_box.text))

    @on(SimpleButton.Pressed, selector='#copy_path')
    def copy_path(self):
        self.app.copy_to_clipboard(self.super_error_dict[1][1][0])  #  [1][1][0] = log file text

    @on(SimpleButton.Pressed, selector='#open_browser')
    def read_report(self):
        self.app.open_browser(self.super_error_dict[1][1][1])   # [1][1][1] = log file html

    @on(SimpleButton.Pressed, selector='#open_github')
    def open_github(self):
        self.app.handle_link(self.gitrepo_path)

    ###~ Keybindings ~###

    def action_focus_previous(self):
        self.focus_previous()
    
    def action_focus_next(self):
        self.focus_next()

    def action_quit(self):
        self.quitme()

    def action_ignore(self):
        self.ignore()


